﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace SqrxzEdit
{
    class Level
    {
        public static readonly int TimeMax = 15000;
        public static readonly int TileSqrxz = 128;
        public static readonly int TileExit = 104;

        public int Length;
        public int Time;
        public int[,] Tiles;
        public Bitmap LevelImage;

        private Bitmap tileset;
        private readonly int[,] tileChart = {{0,1,2,3,12,13,94,15,17,16,18,95,97,109,110,111},
                                             {8,9,52,53,32,33,34,35,64,93,98,65,66,67,68,69},
                                             {10,11,54,55,36,37,38,39,101,102,70,71,72,73,74,75},
                                             {48,49,56,57,0,40,41,0,103,104,76,77,78,79,80,81},
                                             {50,51,58,59,0,42,43,0,105,106,82,83,84,85,86,87},
                                             {46,47,62,63,60,44,45,61,107,108,88,89,90,91,92,0},
                                             {128,129,130,131,132,133,0,0,0,0,0,0,0,0,0,0}};

        public Level()
        {
            Length = 15;
            Time = TimeMax;
            Tiles = new int[8, 255];
            LevelImage = null;
        }


        public Level(int length, int time)
        {
            Length = length;
            Time = time;
            Tiles = new int[8, 255];
            LevelImage = null;
        }


        public void LoadTileData(byte[] data)
        {
            for (int i = 0; i < data.Length; i++)
                Tiles[i % 8, i / 8] = data[i];
        }


        public void Render()
        {
            // load tileset
            System.Reflection.Assembly myAssembly = System.Reflection.Assembly.GetExecutingAssembly();
            Stream myStream = myAssembly.GetManifestResourceStream("SqrxzEdit.sqrxztiles_sml.png");
            tileset = new Bitmap(myStream);

            // draw level
            LevelImage = new Bitmap(4080, 128);
            Graphics g = Graphics.FromImage(LevelImage);
            for (int y = 0; y < 8; y++)
            {
                for (int x = 0; x < 255; x++)
                {
                    int tileNum = Tiles[y, x];
                    Bitmap tileImage = tileset.Clone(new Rectangle((tileNum % 16) * 16, (tileNum / 16) * 16, 16, 16), tileset.PixelFormat);
                    g.DrawImage(tileImage, x * 16, y * 16);
                }
            }
        }


        public byte[] ToArray()
        {
            List<byte> data = new List<byte>();
            for (int x = 0; x < Length; x++)
                for (int y = 0; y < 8; y++)
                    data.Add((byte)Tiles[y, x]);
            return data.ToArray();
        }


        public void ApplyBrush(int x, int y, Rectangle brush)
        {
            for (int by = 0; by < brush.Height; by++)
                for (int bx = 0; bx < brush.Width; bx++)
                    if ((y + by) < 8 && (x + bx) < Length)
                        putTile(y + by, x + bx, tileChart[by + brush.Y, bx + brush.X]);
        }


        private void putTile(int y, int x, int tile)
        {
            if (tile == 3)
            {
                Random r = new Random();
                tile = r.Next(3, 7);
            }
            Tiles[y, x] = tile;
            Graphics g = Graphics.FromImage(LevelImage);
            Bitmap tileImage = tileset.Clone(new Rectangle((tile % 16) * 16, (tile / 16) * 16, 16, 16), tileset.PixelFormat);
            g.DrawImage(tileImage, x * 16, y * 16);
        }


        public void InsertColumn(int x)
        {
            if (Length < 255)
            {
                for (int c = Length; c > x; c--)
                    for (int y = 0; y < 8; y++)
                        putTile(y, c, Tiles[y, c - 1]);
                Length++;
                for (int y = 0; y < 8; y++)
                    putTile(y, x, 0);
            }
        }


        public void DeleteColumn(int x)
        {
            if (Length > 15)
            {
                for (int c = x; c < (Length - 1); c++)
                    for (int y = 0; y < 8; y++)
                        putTile(y, c, Tiles[y, c + 1]);
                Length--;
                for (int y = 0; y < 8; y++)
                    putTile(y, Length, 0);
            }
        }


        public bool Validate()
        {
            int numSqrxz = 0, numExits = 0;
            for (int x = 0; x < Length; x++)
            {
                for (int y = 0; y < 8; y++)
                {
                    if (Tiles[y, x] == TileSqrxz)
                        numSqrxz++;
                    else if (Tiles[y, x] == TileExit)
                        numExits++;
                }
            }
            if (numSqrxz != 1)
            {
                MessageBox.Show("All levels must have exactly one Sqrxz spawn", "Error");
                return false;
            }
            if (numExits == 0)
            {
                MessageBox.Show("All levels must have at least one Exit Door", "Error");
                return false;
            }
            return true;
        }
    }
}
